"""Provides an example for manually serializing data of a plugin that should be shared between
plugin instances and documents.

Entities:
    GetRailSpline:          Returns a rail spline for the given parameters.
    GetProfileSpline:       Returns a profile spline for the given parameters.
    SplinePresetCollection: Abstracts reading and writing spline presets from and to disk.
    StoredProfilesData:     Realizes an object plugin hook that relies for its output on data that 
                            has been serialized to disk.
"""
import c4d

import os
import json
import math
import typing

# --- The generator functions for the core output of the plugin, a rail and profile spline. 

def GetRailSpline(pointCount: int, length: float) -> c4d.SplineObject:
    """Returns a rail spline for the given parameters.
    """
    spline: c4d.SplineObject = c4d.SplineObject(pointCount, c4d.SPLINEOBJECT_TYPE_LINEAR)
    if not spline:
        raise MemoryError("Could not allocate rail spline.")

    # Build the points between 0 and #length with #pointCount steps.
    points: list[c4d.Vector] = [
        c4d.Vector(c4d.utils.RangeMap(i, 0, pointCount - 1, 0, length, True), 0, 0) 
        for i in range(pointCount)]
    
    spline.SetAllPoints(points)
    spline.SetName("Rail")
    spline.Message(c4d.MSG_UPDATE)

    return spline


def GetProfileSpline(pointCount: int, diameter: float) -> c4d.SplineObject:
    """Returns a profile spline for the given parameters.
    """
    spline: c4d.SplineObject = c4d.SplineObject(pointCount, c4d.SPLINEOBJECT_TYPE_LINEAR)
    if not spline:
        raise MemoryError("Could not allocate profile spline.")

    # I saw some "manual" rotations in your code using the sine and cosine. It is a bit a question
    # of taste, but I would recommend using transforms instead. We rotate here with each iteration
    # in the list comprehension the vector #p with a length of #diameter around the z-axis,
    # resulting in a circle (when the splined is closed)
    p: c4d.Vector = c4d.Vector(0, diameter, 0)
    points: list[c4d.Vector] = [
        c4d.utils.MatrixRotZ(2 * math.pi * (float(i)/float(pointCount))) * p
        for i in range(pointCount)]

    spline.SetAllPoints(points)
    spline[c4d.SPLINEOBJECT_CLOSED] = True
    spline.SetName("Profile")
    spline.Message(c4d.MSG_UPDATE)

    return spline

# --- The interface which wraps the whole preset logic. 

class SplinePresetCollection:
    """Abstracts reading and writing spline presets from and to disk.
    
    The interface returns and accepts SplineObject instances via __getitem__ and __setitem__, hiding
    the internal data away for an internal user.

        handler = SplinePresetCollection(somePath, autoSave=True)
        # This will cause #spline to be saved to disk.
        handler["MyPreset"] = spline
        # Load the data back, handler does not store SplineObject instances, but their point data.
        otherSpline: c4d.SplineObject = handler["MyPreset"]

    There many ways how one can realize (de-)serialization of data. HyperFile has the advantage that
    it can natively deal with many Cinema 4D data types. Human readable formats as JSON or XML have
    the advantage that they are modifiable by users without specialized tools.

    I went here with JSON but one could also easily use HyperFile, XML, etc. It is mostly a question
    of taste. I am also serializing point data instead of parameters. Simply because it is the 
    (slightly) more complex thing to do and therefore covers a broader audience.
    """
    # Keys for an item in the internal data structure and the JSON IO.
    KEY_LABEL: str = "label"
    KEY_POINTS: str = "points"

    def __init__(self, path: str, autoSave: bool = True) -> None:
        """Initializes a collection with a preset file at #path to read and write data to.

        When #autosave is #True, the interface will automatically serialize itself when its has 
        state changed. Auto-loading is not required, because all plugin instances share the same
        SplinePresetCollection handler below (but would also be possible if one wants to change
        the class bound SplinePresetCollection instance below).
        """
        if not isinstance(path, str):
            raise TypeError(f"{path = }")
        if not isinstance(autoSave, bool):
            raise TypeError(f"{autoSave = }")
        if not os.path.exists(path):
            raise IOError(f"Profile preset path '{path}' does not exist.")
        if not os.path.isfile(path) or not path.lower().endswith("json"):
            raise IOError(f"Profile preset path '{path}' is not a JSON file.")

        self._path: str = path
        self._autoSave: bool = autoSave

        # The internal data of the collection. We store splines over their label and points in
        # local spline space, e.g.,
        #
        #   [
        #       {"label": "Rectangle", "points": [[-1, 1, 0], [1, 1, 0], [1, -1, 0], [-1, -1, 0]]},
        #       {"label": "Triangle", "points": [[-100, 200, 0], [100, 100, 0], [100, -100, 0]]}
        #   ]
        #
        # We only handle here linearly interpolated splines with no tangents. We could of course
        # also store other parameters than the label of the spline, and then generate the spline
        # over these parameters.
        self._data: list[dict[str, typing.Union[str, list[list[float, int]]]]] = self.Read()
    
    def __len__(self) -> int:
        """Returns the number of presets in the handler instance.
        """
        return len(self._data)

    def __getitem__(self, key: typing.Union[int, str]) -> c4d.SplineObject:
        """Returns a spline object for the preset with the index or label #key.
        """
        if not isinstance(key, (int, str)):
            raise TypeError(f"Invalid key type: {key}.")

        # Get the data.
        pointData: list[list[float, int]] = None
        if isinstance(key, int):
            if key > (len(self._data) - 1):
                raise IndexError(f"The index {key} is out of bounds for {self}.")
            pointData = self._data[key][SplinePresetCollection.KEY_POINTS]
        else:
            for item in self._data:
                if item[SplinePresetCollection.KEY_LABEL] == key:
                    pointData = item[SplinePresetCollection.KEY_POINTS]
        if pointData is None:
            raise KeyError(f"{key} is a key not contained in {self}.")

        points: list[c4d.Vector] = [c4d.Vector(*p) for p in pointData]
        spline: c4d.SplineObject = c4d.SplineObject(len(points), c4d.SPLINEOBJECT_TYPE_LINEAR)
        if not spline:
            raise MemoryError("Failed to allocate spline for preset data.")

        spline[c4d.SPLINEOBJECT_CLOSED] = True
        spline.SetAllPoints(points)
        spline.Message(c4d.MSG_UPDATE)

        return spline

    def __iter__(self) -> tuple[int, str]:
        """Yields the indices and labels of the splines in the collection.

        Point/spline data is not being yielded, use __getitem__ with either the yielded index
        or label for that.
        """
        for i, item in enumerate(self._data):
            yield i, item[SplinePresetCollection.KEY_LABEL]
    
    def AddItem(self, spline: c4d.SplineObject, label: str) -> None:
        """Adds #spline to the data under #label.
        """
        if not isinstance(spline, c4d.SplineObject):
            raise TypeError(f"Invalid item type: {spline = }.")

        self._data.append({
            SplinePresetCollection.KEY_LABEL: label,
            SplinePresetCollection.KEY_POINTS: [[p.x, p.y, p.z] for p in spline.GetAllPoints()]
        })

        if self._autoSave:
            self.Write()

    def Read(self) -> list[dict[str, typing.Union[str, list[list[float, int]]]]]:
        """Reads the preset data from the path associated with this collection.
        """
        # Try to parse the data.
        try:
            with open(self._path, "r") as f:
                data: dict = json.load(f)
        except Exception as e:
            raise IOError(f"Loading profile data failed with the error: {e}")

        # Validate the schema of the data. One could also use JSON schema for that, but I kept it
        # simple here. For more complex cases manual validation is usually not a good choice.
        # HyperFile data would also have to be validated for being in a well-formed state, but
        # there it would be only bugs in the plugin which could cause malformed data, and not also
        # a user errors in manually editing the file.
        if not isinstance(data, list):
            raise IOError(f"Invalid profile data type: {type(data) = }")

        for item in data:
            if not isinstance(item, dict) or len(item) != 2:
                raise TypeError(f"Invalid item type: {item = }")

            if not "label" in item:
                raise KeyError(f"'label' key is missing in : {item = }")
            if not "points" in item:
                raise KeyError(f"'points' key is missing in : {item = }")

            if not isinstance(item["label"], str):
                raise TypeError(f"Invalid label: {item['label'] = }")
            if not isinstance(item["points"], list):
                raise TypeError(f"Invalid points: {item['label'] = }")

            for point in item["points"]:
                if not isinstance(point, list):
                    raise TypeError(f"Invalid point data: {point = }")
                for component in point:
                    if not isinstance(component, (float, int)):
                        raise TypeError(f"Invalid point data: {point = }")
        return data

    def Write(self) -> None:
        """Writes the current state of the collection to disk.
        """
        if not c4d.threading.GeIsMainThreadAndNoDrawThread():
            raise RuntimeError(f"Could not save data from thread: {c4d.threading.GeGetCurrentThread()}")
        try:
            with open(self._path, "w") as f:
                json.dump(self._data, f)
        except Exception as e:
            raise IOError(f"Failed to serialize spline preset data: {e}")


class StoredProfilesData(c4d.plugins.ObjectData):
    """Realizes an object plugin hook that relies for its output on data that has been serialized
    to disk.

    The plugin hook returns an Osweep object as its cache. The rail spline is being built 
    dynamically on each cache build-event. The profiles are wrapped by a "preset" logic which
    writes the profile data as JSON to disk.
    """
    # The plugin ID of StoredProfilesData and the ID for the "Profiles" drop-down parameter,
    # we will need it in #GetDDescription below. 
    ID_PLUGIN: int = 1060702
    DID_PROFILE_PRESETS: c4d.DescID = c4d.DescID(
        c4d.DescLevel(c4d.ID_PROFILE_PRESETS, c4d.DTYPE_LONG, 0))

    # A class bound spline preset handler instance shared by all plugin hook instances. By binding
    # it to the class instead of to each StoredProfilesData hook instance, we ensure that all plugin
    # instances operate on the same data without having to read data back from disk. PRESET_HANDLER
    # acts effectively as a singleton. 
    PRESET_HANDLER: SplinePresetCollection = SplinePresetCollection(
        os.path.join(os.path.dirname(__file__), "profiles.json"))

    def __init__(self):
        """Initializes the plugin hook.
        """
        # We do not need any special cache handling, so we can SetOptimizeCache().
        self.SetOptimizeCache(True)
        super().__init__()

    def Init(self, node: c4d.GeListNode) -> bool:
        """Called by Cinema 4D to initialize the plugin object #node.
        """
        self.InitAttr(node, int, c4d.ID_PROFILE_PRESETS)
        self.InitAttr(node, float, c4d.ID_PROFILE_DIAMETER)
        self.InitAttr(node, int, c4d.ID_PROFILE_SUBDIVISIONS)
        self.InitAttr(node, float, c4d.ID_RAIL_LENGTH)
        self.InitAttr(node, int, c4d.ID_RAIL_SUBDIVISIONS)

        node[c4d.ID_PROFILE_PRESETS] = 0
        node[c4d.ID_PROFILE_DIAMETER] = 50.0
        node[c4d.ID_PROFILE_SUBDIVISIONS] = 12
        node[c4d.ID_RAIL_LENGTH] = 500.0
        node[c4d.ID_RAIL_SUBDIVISIONS] = 24

        return True

    def GetVirtualObjects(self, op: c4d.BaseObject, hh: object) -> typing.Optional[c4d.BaseObject]:
        """Called by Cinema 4D to build the cache for the plugin node #op.
        """
        # When you invoke SetOptimizeCache(True) in ObjectData.__Init__(), also applying the default
        # manual cache handling as shown below and done in your code makes only little sense. 
        # Because that is then already being done by Cinema 4D for you. When one is determining the
        # dirty state manually, one usually also wants to deviate from the default behaviour by for 
        # example updating the cache when an object linked in the parameters of #op is dirty, 
        # when the current time of the document has changed, or something similarly custom.
        #
        # # Determine if the object is dirty, and if not, simply return its existing cache. Doing
        # # this is equivalent to invoking self.SetOptimizeCache(True) in __init__().
        # dirty = op.CheckCache(hierarchyhelp) or op.IsDirty(c4d.DIRTYFLAGS_DATA)
        # if dirty is False: return op.GetCache(hierarchyhelp)

        # Get relevant parameters from the node in the scene.
        railLength: float = op[c4d.ID_RAIL_LENGTH]
        railSubdivisions: int = op[c4d.ID_RAIL_SUBDIVISIONS]
        presetIndex: int = op[c4d.ID_PROFILE_PRESETS]

        # Allocate the cache root, in this case a sweep object we are going to attach our customly
        # generated rail and profile splines to.
        root: c4d.BaseObject = c4d.BaseObject(c4d.Osweep)
        if not root:
            raise MemoryError("Could not allocate object in cache.")

        # Get profile spline from the preset handler and build the rail manually.
        profileSpline: c4d.SplineObject = StoredProfilesData.PRESET_HANDLER[presetIndex]
        railSpline: c4d.SplineObject = GetRailSpline(railSubdivisions, railLength)
        profileSpline.InsertUnderLast(root)
        railSpline.InsertUnderLast(root)

        # Return the new cache hierarchy.
        return root

    def GetDDescription(self, node: c4d.GeListNode, description: c4d.Description,
                        flags: int) -> typing.Union[bool, tuple[bool, int]]:
        """Called by Cinema 4D to let a plugin modify its description GUI.

        We use this here to populate our "Profiles" parameter drop-down menu with the data wrapped
        by #StoredProfilesData.PRESET_HANDLER.
        """
        # Make sure that the description of this plugin type has been loaded.
        if not description.LoadDescription(node.GetType()):
            return False

        # Only modify the #ID_PROFILE_PRESETS cycle when Cinema 4D asks us to do so, or when 
        # Cinema 4D does not specify for which parameter this call is for.
        targetId: c4d.DescID = description.GetSingleDescID()
        if not targetId or StoredProfilesData.DID_PROFILE_PRESETS.IsPartOf(targetId):
            # Get the whole description container instance for the parameter #ID_PROFILE_PRESETS.
            bc: c4d.BaseContainer = description.GetParameterI(StoredProfilesData.DID_PROFILE_PRESETS)
            if not bc:
                return (True, flags)
            
            # Get the cycle values container instance in it and flush all items.
            items: c4d.BaseContainer = bc.GetContainerInstance(c4d.DESC_CYCLE)
            items.FlushAll()

            # Build the values from scratch with the items in #StoredProfilesData.PRESET_HANDLER.
            for index, label in StoredProfilesData.PRESET_HANDLER:
                items.SetString(index, label)

        return (True, flags | c4d.DESCFLAGS_DESC_LOADED)
    
    def Message(self, node: c4d.GeListNode, mid: int, data: object) -> bool:
        """Called by Cinema 4D to convey events for the plugin object #node.

        Used here to catch the user pressing the "Save Preset" button and saving said preset.
        """
        # Our #ID_PROFILE_SAVE button has been pressed, we save a new preset.
        if (mid == c4d.MSG_DESCRIPTION_COMMAND and isinstance(data, dict) and
            isinstance(data.get("id", None), c4d.DescID)):
            if data["id"][0].id != c4d.ID_PROFILE_SAVE:
                return True
            
            # Bail when we are not on the non drawing main thread as we are going to open a dialog
            # to ask the user for a preset label.
            if not c4d.threading.GeIsMainThreadAndNoDrawThread():
                return True
            label: str = c4d.gui.InputDialog(
                "Preset Name:", f"Preset {len(StoredProfilesData.PRESET_HANDLER) + 1}")

            # Build the profile spline with the current settings.
            diameter: float = node[c4d.ID_PROFILE_DIAMETER]
            subdivisions: int = node[c4d.ID_PROFILE_SUBDIVISIONS]
            spline: c4d.SplineObject = GetProfileSpline(subdivisions, diameter)

            # Store the spline in the handler, because we set #autoSave to #True, this will also 
            # cause the data to be saved to disk. After that, we flag ourselves as description dirty,
            # as the drop-down menu of the "Preset" parameter must be updated.
            StoredProfilesData.PRESET_HANDLER.AddItem(spline, label)
            node.SetDirty(c4d.DIRTYFLAGS_DESCRIPTION)

        return True

    @staticmethod
    def Register() -> bool:
        """Registers the plugin hook.

        This method is not part of the ObjectData interface, I just like to do it like this, attach
        the plugin ID and registration method to the plugin hook they belong to. Could also be
        done differently.
        """
        bmp: c4d.bitmaps.BaseBitmap = c4d.bitmaps.InitResourceBitmap(c4d.Osweep)
        if not bmp:
            return False

        return c4d.plugins.RegisterObjectPlugin(id=StoredProfilesData.ID_PLUGIN,
                                                str="Stored Profiles Object",
                                                g=StoredProfilesData,
                                                description="ostoredprofiles",
                                                icon=bmp, info=c4d.OBJECT_GENERATOR)


if __name__ == "__main__":
    if not StoredProfilesData.Register():
        print(f"Warning: Failed to register 'StoredProfilesData' in '{__file__}'.")
